!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_do_ordering.F90
!! This file contains the routine computes (through different methods) a column permutation
!! of the input matrix in order to reduce fill-in
!!
!!
!! ##############################################################################################


#include "qrm_common.h"


!> @brief This routine computes (through different methods) a column permutation
!! of the input matrix in order to reduce fill-in

!> Supported methods are, currently, COLAMD, SCOTCH and METIS. The user may also
!! provide his own permutation, in which case a check on its validity is done.
!!
!! @param[in] graph    the graph associated to the matrix to be ordered.
!!
!! @param[out] cperm   the new column order
!!
!! @param[in] cperm_in the permutation eventually provided by the user
!!
subroutine dqrm_do_ordering(graph, qrm_spfct, cperm, cperm_in, info)

  use dqrm_spmat_mod
  use dqrm_spfct_mod
  use qrm_error_mod
  use dqrm_analysis_mod, savesym => dqrm_do_ordering
  use qrm_parameters_mod
  use qrm_common_mod
  implicit none

  type(dqrm_spmat_type), target  :: graph
  type(dqrm_spfct_type)          :: qrm_spfct
  integer                        :: cperm(:)
  integer, pointer               :: cperm_in(:)
  integer, optional              :: info

  integer                        :: iord=0
  integer                        :: i, ts, te, tr, cnt
  type(dqrm_spmat_type), target  :: graph_pruned
  type(dqrm_spmat_type), pointer :: graph_ptr
  integer, pointer               :: tag(:)
  
  ! error management
  integer                        :: err
  character(len=*), parameter    :: name='qrm_do_ordering'

  err = 0

  call qrm_get(qrm_spfct, 'qrm_ordering', iord)
  if(iord .eq. qrm_auto_) then
     iord = qrm_choose_ordering()
  end if

  if(qrm_spfct%icntl(qrm_schur_).gt.0) then
     call dqrm_spmat_prune(graph, graph_pruned, &
          qrm_spfct%cperm_in, qrm_spfct%icntl(qrm_schur_))
     graph_ptr => graph_pruned
  else
     graph_ptr => graph
  end if
  
  select case(iord)
  case(qrm_natural_)
     ! perm has to be allocated and built explicitly because
     ! we are going to use an equivalent permutation
     do i=1, graph_ptr%n
        cperm(i) = i
     end do
  case(qrm_given_)
     ! in this case we just need to check that
     ! the given permutation makes sense
     if(.not. associated(cperm_in)) then
        err = 8; call qrm_error_print(err, name)
        goto 9999
     else
        call qrm_check_cperm(cperm_in, graph_ptr%n, err)
        __QRM_INFO_CHECK(err, name,'qrm_check_perm',9999)
        cperm(1:graph_ptr%n) = cperm_in(1:graph_ptr%n)
     end if
  case(qrm_colamd_)
     ! COLAMD
#if defined (have_colamd)
     call dqrm_do_colamd(graph_ptr, cperm, err)
     __QRM_INFO_CHECK(err, name,'qrm_do_colamd',9999)
#else
     err=16; call qrm_error_print(err, name, aed='colamd')
     goto 9999
#endif
  case(qrm_metis_)
     ! METIS
#if defined (have_metis)
     call dqrm_do_metis(graph_ptr, cperm, err)
     __QRM_INFO_CHECK(err, name,'qrm_do_metis',9999)
#else
     err=16; call qrm_error_print(err, name, aed='metis')
     goto 9999
#endif
  case(qrm_scotch_)
     ! SCOTCH
#if defined (have_scotch)
     call dqrm_do_scotch(graph_ptr, cperm, err)
     __QRM_INFO_CHECK(err, name,'qrm_do_scotch',9999)
#else
     err=16; call qrm_error_print(err, name, aed='scotch')
     goto 9999
#endif
  case default
     err=9; call qrm_error_print(err, name, ied=(/iord/))
     goto 9999
  end select

  if(qrm_spfct%icntl(qrm_schur_).gt.0) then
     ! push schur complement variables to the end
     ! use graph_pruned%jptr as a tagging vector
     tag => graph_pruned%jptr
     tag(qrm_spfct%cperm_in(1:qrm_spfct%icntl(qrm_schur_))) = -1
     cnt=0
     do i=1, graph%n
        if(tag(cperm(i)).ne.-1) then
           cnt = cnt+1
           cperm(cnt)=cperm(i)
        end if
     end do
     cperm(cnt+1:graph%n) = qrm_spfct%cperm_in(1:qrm_spfct%icntl(qrm_schur_))
     call qrm_spmat_destroy(graph_pruned)
  end if

  
9999 continue 
  
  if(present(info)) info=err
  return

contains

  function qrm_choose_ordering()
    ! Function: qrm_choose_ordering
    ! This function sets the ordering for the case where an automatic
    ! choice is requested
    !

    integer :: qrm_choose_ordering
    integer :: iord

    iord = 1

#if defined(have_colamd)
    iord = qrm_colamd_
#endif

#if defined(have_scotch)    
    iord = qrm_scotch_
#endif

#if defined(have_metis)    
    iord = qrm_metis_
#endif

    qrm_choose_ordering = iord
    return
    
  end function qrm_choose_ordering


end subroutine dqrm_do_ordering
