!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing the QR factorization of a triangle on top of a rectangle
!! or triangle
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine dqrm_dsmat_tpqr(a, b, t, ib, work, m, n, l, prio, info)

  use dqrm_dsmat_mod, alias=>dqrm_dsmat_tpqr
  use qrm_dscr_mod
  implicit none

  type(dqrm_dsmat_type)       :: a, b, t
  type(dqrm_ws_type)          :: work
  integer                     :: ib
  integer, optional           :: m, n, l, prio
  integer, optional           :: info

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_tpqr'

  err = 0
  call qrm_dscr_init(qrm_dscr)

  call dqrm_dsmat_tpqr_async(qrm_dscr, a, b, t, ib, work, m, n, l, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return
end subroutine dqrm_dsmat_tpqr

subroutine dqrm_dsmat_tpqr_async(qrm_dscr, a, b, t, ib, work, m, n, l, prio)

  use dqrm_dsmat_mod, alias=>dqrm_dsmat_tpqr_async
  use qrm_dscr_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(dqrm_dsmat_type)       :: a, b, t
  type(dqrm_ws_type)          :: work
  integer                     :: ib
  integer, optional           :: m, n, l, prio

  integer                     :: i, j, im, in, il, p, nbm, nbl, nbe
  integer                     :: mm, nn, kk, ll
  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_tpqr_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  if(min(a%m,a%n).eq.0) return

  __QRM_IFPRESENT3(im, m, b%m)
  __QRM_IFPRESENT3(in, n, a%n)
  __QRM_IFPRESENT3(il, l, 0)

  nbe = dqrm_dsmat_inblock(a, in)
  nbm = dqrm_dsmat_inblock(b, im)
  nbl = dqrm_dsmat_inblock(b, il)

  do p=1, nbe
     kk = min(a%f(p+1)-a%f(p), in-a%f(p)+1)
     do i = 1, min(nbm, nbm-nbl+p)
        mm = min(b%f(i+1)-b%f(i), im-b%f(i)+1)
        ! ll = max((i-1)*b%mb + mm - (im-il+ (p-1)*b%mb),0)
        ll = max(b%f(i)     + mm - (im-il+ b%f(p))    , 0)
        call dqrm_hitpqrt(qrm_dscr,           &
             mm, kk, ll, max(mm,kk), ib,      &
             a%blocks(p,p),                   &
             b%blocks(i,p),                   &
             t%blocks(i,p),                   &
             work, prio)
        do j=p+1, nbe
           ! nn = min(in-(j-1)*a%mb,a%mb)
           nn = min(a%f(j+1)-a%f(j), in-a%f(j)+1)
           call dqrm_hitpmqrt(qrm_dscr,         &
                mm, nn, kk, ll, max(kk,nn), ib, &
                b%blocks(i,p),                  &
                t%blocks(i,p),                  &
                a%blocks(p,j),                  &
                b%blocks(i,j),                  &
                work, prio)
        end do
     end do

  end do

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine dqrm_dsmat_tpqr_async



! subroutine dqrm_dsmat_init_t_tpqr(qrm_dsmat_a, qrm_dsmat_t, ib, m, n, l, pin, seq, info)

! #if defined (have_starpu)
!   use fstarpu_mod
!   use qrm_starpu_common_mod
! #endif
!   use qrm_parameters_mod
!   use qrm_error_mod
!   use qrm_mem_mod
!   use dqrm_dsmat_mod, alias=>dqrm_dsmat_init_t_tpqr
!   use qrm_dscr_mod

!   implicit none

!   type(dqrm_dsmat_type), intent(in)            :: qrm_dsmat_a
!   type(dqrm_dsmat_type), intent(inout), target :: qrm_dsmat_t
!   integer, intent(in)                          :: ib
!   integer, optional                            :: m, n, l, info
!   logical, optional                            :: pin, seq

!   integer                                      :: i, j, is, js, parts, nbr, nbc, nb, im, in, il
!   type(c_ptr)                                  :: bc_filter
!   logical                                      :: iseq
  
!   ! error handling
!   integer                                      :: err
!   character(len=*), parameter                  :: name='qrm_dsmat_init_t_tpqr'

!   err = 0

!   if(.not.qrm_dsmat_a%inited) return
  
!   __QRM_IFPRESENT3(iseq, seq, .false.)
!   __QRM_IFPRESENT3(im, m, a%m)
!   __QRM_IFPRESENT3(in, n, a%n)
!   __QRM_IFPRESENT3(il, l, 0)

!   nbe = (in-1)/a%mb+1
!   nbm = (im-1)/a%mb+1
!   nbl = (il-1)/a%mb+1
  
!   ! nbr = size(qrm_dsmat_a%blocks,1)
!   ! nbc = size(qrm_dsmat_a%blocks,2)
!   nb  = qrm_dsmat_a%nb

!   ! FIXME: add error checks

! #if defined (have_starpu)
!   if(.not.iseq) then
!      bc_filter = qrm_get_bc_filter()
!      call fstarpu_data_filter_set_filter_arg (bc_filter, nb)
!   end if
! #endif

!   allocate(qrm_dsmat_t%blocks(nbm,nbe))

!   do p=1, nbe
!      kk = min(a%mb,in-(p-1)*a%mb)

!      do i=1, min(nbm, nbm-nbl+p)
!         if(.not.qrm_allocated(qrm_dsmat_a%blocks(i,p)%c)) cycle

!         mm = min(b%mb, im-(i-1)*b%mb)

!         ! is = size(qrm_dsmat_a%blocks(i,p)%c,1)
!         ! js = size(qrm_dsmat_a%blocks(i,p)%c,2)

!         call qrm_alloc(qrm_dsmat_t%blocks(i,p)%c, ib, kk, err, pin)
!         __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

!         qrm_dsmat_t%blocks(i,p)%c = qrm_dzero

! #if defined (have_starpu)
!         if(.not.iseq) then
!            ! register the block to starpu
!            call fstarpu_matrix_data_register(qrm_dsmat_t%blocks(i,p)%hdl, 0, &
!                 c_loc(qrm_dsmat_t%blocks(i,p)%c),                            &
!                 size(qrm_dsmat_t%blocks(i,p)%c,1),                           &
!                 size(qrm_dsmat_t%blocks(i,p)%c,1),                           &
!                 size(qrm_dsmat_t%blocks(i,p)%c,2),                           &
!                 c_sizeof(qrm_dsmat_t%blocks(i,p)%c(1,1)))
!            if(nb.lt.js) then
!               ! plan a dynamic partitioning
!               parts = (kk-1)/nb+1
!               allocate(qrm_dsmat_t%blocks(i,p)%shdls(parts))
!               call fstarpu_data_partition_plan(qrm_dsmat_t%blocks(i,p)%hdl, bc_filter, qrm_dsmat_t%blocks(i,p)%shdls)
!            end if

!         end if
! #endif
!         qrm_dsmat_t%blocks(i,p)%partitioned = .false.
!      end do
!   end do

!   qrm_dsmat_t%inited = .true.

! 9999 continue 
!   ! should do cleanup
!   if(present(info)) info = err
!   return
! end subroutine dqrm_dsmat_init_t_tpqr
