!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! filling a matrix
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine sqrm_dsmat_fill(a, init, ivalue, i, j, m, n, l, prio, info)

  use sqrm_dsmat_mod, alias=>sqrm_dsmat_fill
  use qrm_dscr_mod

  implicit none

  type(sqrm_dsmat_type), target   :: a
  integer, optional               :: i, j, m, n, l
  character                       :: init
  real(r32), optional          :: ivalue 
  integer, optional               :: prio, info
                                   
  type(qrm_dscr_type)             :: qrm_dscr
                                   
  ! error handling                 
  integer                         :: err
  character(len=*), parameter     :: name='qrm_dsmat_fill'

  err = 0

  if(.not.a%inited) then
     err = 1000
     call qrm_error_print(err, name)
     goto 9999
  end if

  call qrm_dscr_init(qrm_dscr, nocuda=.true.)
  call sqrm_dsmat_fill_async(qrm_dscr, a, init, ivalue, i, j, m, n, l, prio)
  call qrm_barrier(qrm_dscr)
  err = qrm_dscr%info

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return

end subroutine sqrm_dsmat_fill

subroutine sqrm_dsmat_fill_async(qrm_dscr, a, init, ivalue, i, j, m, n, l, prio)
  use sqrm_dsmat_mod, s=>sqrm_dsmat_fill_async
  use qrm_dscr_mod
  use sqrm_dsmat_task_mod
  use qrm_string_mod
  use qrm_parameters_mod
#if defined (have_starpu)
  use sqrm_starpu_mod
  use fstarpu_mod
#endif
  
  implicit none

  type(qrm_dscr_type)            :: qrm_dscr
  type(sqrm_dsmat_type), target  :: a
  character                      :: init
  integer, optional              :: i, j, m, n, l, prio
  real(r32), optional         :: ivalue 

  type(sqrm_block_type), pointer :: aij
  integer                        :: mb, nb, iprio
  integer                        :: ii, ij, im, in, il, mm, nn
  real(r32)                   :: iivalue 
  integer                        :: first_rblock, first_cblock
  integer                        :: last_rblock, last_cblock
  integer                        :: br, bc, bi, bj, bm, bn, bl
  
  ! error handling
  integer                        :: info, err
  character(len=*), parameter    :: name='qrm_dsmat_fill_async'


  if(qrm_dscr%info.ne.0) return
  info = 0

  if(.not.a%inited) then
     info = 1000
     call qrm_error_print(info, name)
     goto 9999
  end if

  __QRM_IFPRESENT3(ii, i, 1)
  __QRM_IFPRESENT3(ij, j, 1)
  __QRM_IFPRESENT3(im, m, a%m-ii+1)
  __QRM_IFPRESENT3(in, n, a%n-ij+1)
  __QRM_IFPRESENT3(il, l, 0)
  __QRM_IFPRESENT3(iivalue, ivalue, qrm_szero)

  first_rblock = sqrm_dsmat_inblock(a, ii)
  first_cblock = sqrm_dsmat_inblock(a, ij)
  last_rblock  = sqrm_dsmat_inblock(a, ii+im-1)
  last_cblock  = sqrm_dsmat_inblock(a, ij+in-1)

  do bc = first_cblock, last_cblock
     do br = first_rblock, last_rblock
        call sqrm_dsmat_block_ijmnl(a, ii, ij, im, in, il, br, bc, bi, bj, bm, bn, bl)
        if(min(bm,bn).gt.0) then
           aij => a%blocks(br,bc)
           if(qrm_allocated(aij%c)) &
                call sqrm_block_fill_task(qrm_dscr, &
                     aij,                           &
                     bi, bj,                        &
                     bm, bn, bl,                    &
                     init, iivalue, prio)
        end if
     end do
  end do

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine sqrm_dsmat_fill_async
