!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing a triangular system solution
!!
!! Assume A=[A1 A2] where A1 is triangular of size mxm and A2 is of
!! size mx(k-m) and B=[B1; B2]:
!!
!! if trans='n'
!!    B1 = A1\(B1-A2*B2)
!!
!! if trans='t'
!!    B1 += A1'\B1   B2 -= A2'(A1'\B1) 
!!
!! ##############################################################################################

#include "qrm_common.h"


subroutine sqrm_dsmat_trsm(side, uplo, trans, diag, alpha, a, b, m, n, k, prio, info)

  use sqrm_dsmat_mod, alias => sqrm_dsmat_trsm
  use qrm_dscr_mod
  implicit none

  type(sqrm_dsmat_type)       :: a, b
  real(r32)                :: alpha
  character                   :: side, uplo, trans, diag
  integer, optional           :: m, n, k, prio, info

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_trsm'

  err = 0

  call qrm_dscr_init(qrm_dscr)

  call sqrm_dsmat_trsm_async(qrm_dscr, side, uplo, trans, diag, alpha, a, b, m, n, k, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return

end subroutine sqrm_dsmat_trsm

subroutine sqrm_dsmat_trsm_async(qrm_dscr, side, uplo, trans, diag, alpha, a, b, m, n, k, prio)

  use sqrm_dsmat_mod, alias => sqrm_dsmat_trsm_async
  use qrm_dscr_mod
  use sqrm_dsmat_task_mod
  use qrm_parameters_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(sqrm_dsmat_type)       :: a, b
  real(r32)                :: alpha
  character                   :: side, uplo, trans, diag
  integer, optional           :: prio, m, n, k

  integer                     :: i, j, l
  real(r32)                :: ialpha
  integer                     :: im, in, ik, mm, nn, kk
  integer                     :: last_mblock, last_nblock, last_kblock

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_trsm_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  __QRM_IFPRESENT3(im, m, a%m)
  __QRM_IFPRESENT3(in, n, b%n)
  __QRM_IFPRESENT3(ik, k, a%n)

  im = min(im,ik)

  if(minval((/im,in,ik/)).le.0) return
  
  last_mblock = sqrm_dsmat_inblock(a, im)
  last_nblock = sqrm_dsmat_inblock(b, in)
  last_kblock = sqrm_dsmat_inblock(a, ik)

  if(side.eq.qrm_left) then
     if(uplo.eq.qrm_upper) then
        if(trans.eq.qrm_no_transp) then
           do j=1, last_nblock
              if(j.eq.last_nblock) then
                 nn = in-b%f(j)+1
              else
                 nn = b%f(j+1)-b%f(j)
              end if
              
              do l = last_kblock, 1, -1
                 if(l.eq.last_kblock) then
                    kk = ik-a%f(l)+1
                    ialpha = alpha
                 else
                    kk = a%f(l+1)-a%f(l)
                    ialpha = qrm_sone
                 end if

                 ! b_j = ialpha*b_j/a_jj
                 if(l.le.last_mblock) then
                    if(l.eq.last_mblock) then
                       mm = im-a%f(l)+1
                    else
                       mm = a%f(l+1)-a%f(l)
                    end if
                    call sqrm_trsm_task(qrm_dscr, &
                         qrm_left, qrm_upper,     &
                         qrm_no_transp, diag,     & 
                         mm, nn, kk,              &
                         ialpha,                  & 
                         a%blocks(l,l),           & 
                         b%blocks(l,j),           & 
                         prio )
                 end if
                 do i=min(l-1, last_mblock), 1, -1
                    if(i.eq.last_mblock) then
                       mm = im-a%f(i)+1
                    else
                       mm = a%f(i+1)-a%f(i)
                    end if
                    call sqrm_gemm_task(qrm_dscr,      &
                         qrm_no_transp, qrm_no_transp, &
                         mm, nn, kk,                   &
                         qrm_smone,                    &
                         a%blocks(i,l),                &
                         b%blocks(l,j),                &
                         ialpha,                       &
                         b%blocks(i,j),                &
                         prio)
                 end do

              end do
           end do
        else if((trans.eq.qrm_transp).or.(trans.eq.qrm_transp)) then
           do j=1, last_nblock
              if(j.eq.last_nblock) then
                 nn = in-b%f(j)+1
              else
                 nn = b%f(j+1)-b%f(j)
              end if

              do i= 1, min(last_mblock,last_kblock)
                 if(i.eq.1) then
                    ialpha = alpha
                 else
                    ialpha = qrm_sone
                 end if

                 if(i.eq.last_mblock) then
                    mm = im-a%f(i)+1
                 else
                    mm = a%f(i+1)-a%f(i)
                 end if

                 if(i.eq.last_kblock) then
                    kk = ik-a%f(i)+1
                 else
                    kk = a%f(i+1)-a%f(i)
                 end if

                 ! b_i = ialpha*b_i/a_ii'
                 call sqrm_trsm_task(qrm_dscr, &
                      qrm_left, qrm_upper,     &
                      trans, diag,             &   
                      mm, nn, kk,              &
                      ialpha,                  &
                      a%blocks(i,i),           &
                      b%blocks(i,j),           &
                      prio)

                 do l=i+1, last_kblock
                    if(l.eq.last_kblock) then
                       kk = ik-a%f(l)+1
                       ialpha = alpha
                    else
                       kk = a%f(l+1)-a%f(l)
                       ialpha = qrm_sone
                    end if

                    ! b_j = ialpha*b_j - a_ij'*b_i
                    call sqrm_gemm_task(qrm_dscr,        &
                         trans, qrm_no_transp,           &
                         kk, nn, mm,                     &
                         qrm_smone,                      &
                         a%blocks(i,l),                  &
                         b%blocks(i,j),                  &
                         ialpha,                         &
                         b%blocks(l,j),                  &
                         prio )
                 end do
              end do
           end do
        end if
     else
        write(*,'("TRSM with uplo=l not yet implemented")')
     end if
  else
     write(*,'("TRSM with side=r not yet implemented")')
  end if


9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine sqrm_dsmat_trsm_async

