!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! making the sum of two matrices
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine zqrm_dsmat_axpy(a, b, ia, ja, ib, jb, m, n, l, alpha, prio, info)

  use zqrm_dsmat_mod, alias=>zqrm_dsmat_axpy
  use qrm_dscr_mod

  implicit none


  type(zqrm_dsmat_type)          :: a                    ! This is the source
                                                         ! matrix.

  type(zqrm_dsmat_type)          :: b                    ! This is the destination
                                                         ! matrix.

  integer, optional              :: ia, ja, ib, jb       ! ia and ja are the staring row and col
                                                         ! in the A matrix matrix. ib and jb
                                                         ! the same for B
  integer, optional              :: m, n, l              ! m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
  complex(r64), optional         :: alpha                ! alpha scalar
  integer, optional              :: prio, info

  type(qrm_dscr_type)            :: qrm_dscr

  ! error handling
  integer                        :: err
  character(len=*), parameter    :: name='qrm_dsmat_axpy'

  err = 0

  if(.not.b%inited) then
     err = 1000
     call qrm_error_print(err, name)
     goto 9999
  end if

  call qrm_dscr_init(qrm_dscr, nocuda=.true.)

  call zqrm_dsmat_axpy_async(qrm_dscr, a, b, ia=ia, ja=ja, ib=ib, jb=jb, m=m, n=n, l=l, alpha=alpha, prio=prio)
  
  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return

end subroutine zqrm_dsmat_axpy


subroutine zqrm_dsmat_axpy_async(qrm_dscr, a, b, ia, ja, ib, jb, m, n, l, alpha, prio)
  use zqrm_dsmat_mod, alias=>zqrm_dsmat_axpy_async
  use qrm_dscr_mod
  use qrm_mem_mod
  use iso_c_binding
  use zqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type), target    :: qrm_dscr
  type(zqrm_dsmat_type)          :: a                    ! This is the source
                                                         ! matrix.

  type(zqrm_dsmat_type)          :: b                    ! This is the destination
                                                         ! matrix.

  integer, optional              :: ia, ja, ib, jb       ! ia and ja are the staring row and col
                                                         ! in the A matrix matrix. ib and jb
                                                         ! the same for B
  integer, optional              :: m, n, l              ! m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
  complex(r64), optional         :: alpha                ! alpha scalar
  integer, optional              :: prio
  
  integer   :: a_m, a_n, a_mb
  integer   :: b_m, b_n, b_mb
  integer   :: a_i, a_j
  integer   :: b_i, b_j
  integer   :: im, in, il
  integer   :: a_r, a_c
  integer   :: b_r, b_c
  integer   :: rb, cb
  integer   :: a_bi, a_bj, a_bm, a_bn, a_bl, a_ii, a_jj, a_bii, a_bjj
  integer   :: b_bi, b_bj, b_bm, b_bn, b_bl, b_ii, b_jj
  integer   :: bc_i, bc_j, bc_m, bc_n, bc_l
  integer   :: first_arblock, last_arblock
  integer   :: first_acblock, last_acblock
  integer   :: first_brblock, last_brblock
  integer   :: first_bcblock, last_bcblock
  integer   :: off, r, c, tmp
  complex(r64) :: ialpha
  
  ! error handling
  integer                        :: info
  character(len=*), parameter    :: name='qrm_dsmat_axpy_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  __QRM_IFPRESENT3(a_i,ia,1)
  __QRM_IFPRESENT3(a_j,ja,1)
  __QRM_IFPRESENT3(b_i,ib,1)
  __QRM_IFPRESENT3(b_j,jb,1)
  __QRM_IFPRESENT3(im,m,min(b%m-b_i+1,a%m-a_i+1)) 
  __QRM_IFPRESENT3(in,n,min(b%n-b_j+1,a%n-a_j+1)) 
  ! __QRM_IFPRESENT3(im,m,a%m-a_i+1)
  ! __QRM_IFPRESENT3(in,n,a%n-a_j+1)
  __QRM_IFPRESENT3(il,l,0)
  __QRM_IFPRESENT3(ialpha,alpha,qrm_zone)

  if(min(im,in).le.0) return

  if(.not.b%inited) then
     info = 1000
     call qrm_error_print(info, name)
     goto 9999
  end if

  first_arblock = zqrm_dsmat_inblock(a, a_i)
  first_acblock = zqrm_dsmat_inblock(a, a_j)
  last_arblock  = zqrm_dsmat_inblock(a, a_i+im-1)
  last_acblock  = zqrm_dsmat_inblock(a, a_j+in-1)
  
  do a_r=first_arblock, last_arblock
     do a_c=first_acblock, last_acblock
        
        call zqrm_dsmat_block_ijmnl(a,      &
             a_i, a_j, im, in, il,          &
             a_r, a_c,                      &
             a_bi, a_bj, a_bm, a_bn, a_bl)
        
        a_ii = a%f(a_r)+a_bi-1 ! the index of the first row of a(a_r,a_c) in a
        a_jj = a%f(a_c)+a_bj-1 ! the index of the first col of a(a_r,a_c) in a
        
           ! a(a_r,a_c) may span multiple blocks of b
        first_brblock = zqrm_dsmat_inblock(b, b_i+a_ii-a_i)
        first_bcblock = zqrm_dsmat_inblock(b, b_j+a_jj-a_j)
        last_brblock  = zqrm_dsmat_inblock(b, b_i+a_ii-a_i+a_bm-1)
        last_bcblock  = zqrm_dsmat_inblock(b, b_j+a_jj-a_i+a_bn-1)
        ! write(*,'("ar,ac:",i2,x,i2,"  m,n,l:",i2,x,i2,x,i2,"  fr,lr:",i2,x,i2,"  fc,lc:",i2,x,i2)')&
        ! a_r, a_c, a_bm, a_bn, a_bl, &
        ! first_brblock, last_brblock, &
        ! first_bcblock, last_bcblock
        
        ! loop over the blocks of b
        do b_r=first_brblock, last_brblock
           do b_c=first_bcblock, last_bcblock
              
              call zqrm_dsmat_block_ijmnl(b,      &
                   b_i+a_ii-a_i, b_j+a_jj-a_j, a_bm, a_bn, a_bl,    &
                   b_r, b_c,                      &
                   b_bi, b_bj, b_bm, b_bn, b_bl)
              b_ii = b%f(b_r)+b_bi-1 ! the index of the first row of b(b_r,b_c) in b
              b_jj = b%f(b_c)+b_bj-1 ! the index of the first col of b(b_r,b_c) in b
              
              ! a_bii = a_bi + b_ii - b_i+a_ii-a_i
              ! a_bjj = a_bj + b_jj - b_j+a_jj-a_j
              a_bii = (b_ii-b_i+a_i)-a%f(a_r)+1 
              a_bjj = (b_jj-b_j+a_j)-a%f(a_c)+1 
              
              ! write(*,'("   br,bc:",i2,x,i2)', advance='no') b_r,b_c
              ! write(*,'("   bi,bj:",i2,x,i2)', advance='no') b_bi, b_bj
              ! write(*,'("   ai,aj:",i2,x,i2)', advance='no') a_bii, a_bjj
              ! write(*,'("   m,n,l:",i2,x,i2,x,i2)') b_bm, b_bn, b_bl
              
              if(qrm_allocated(a%blocks(a_r,a_c)%c) .and. &
                   qrm_allocated(b%blocks(b_r,b_c)%c)) &
                   call zqrm_block_axpy_task(qrm_dscr, ialpha, &
                   a%blocks(a_r,a_c), b%blocks(b_r,b_c),   &
                   ia=a_bii, ja=a_bjj, ib=b_bi, jb=b_bj,     &
                   m=b_bm, n=b_bn, l=b_bl, prio=prio)
           end do
        end do
     end do
  end do
  
9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine zqrm_dsmat_axpy_async

