!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine cqrm_hitpqrt_task(qrm_dscr, m, n, l, nb, ib, i, a1, a2, t, work, prio)

#if defined(have_starpu)
  use cqrm_starpu_mod
  use qrm_starpu_common_mod
#endif
  use cqrm_dsmat_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(qrm_dscr_type), target   :: qrm_dscr
  type(cqrm_block_type), target :: a1, a2, t
  integer, target               :: m, n, l, nb, ib, i
  type(cqrm_ws_type)            :: work
  integer, optional             :: prio

#if defined(have_starpu)
  type(c_ptr), target           :: dscrpt, strpt
  type(c_ptr)                   :: hdla1, hdla2, hdlt
  real(c_float),target         :: flops
#endif
  integer, target               :: ii, j, e, iprio, na1, lda1, lda2, nt, ldt, info
  integer(kind=8)               :: tb, te
  integer                       :: pr=0
  
  if(qrm_dscr%info.ne.0) return

  if(a1%partitioned) then
     ii     = (i-1)*nb+1
  else
     ii     = 1
  end if


#if defined (have_starpu)

  if(qrm_dscr%seq.eq.0) then

     ! submit a tpqrt task
     if(a1%partitioned) then
        hdla1 = a1%shdls(i)
        hdla2 = a2%shdls(i)
        hdlt  = t%shdls(i)
     else
        hdla1 = a1%hdl
        hdla2 = a2%hdl
        hdlt  = t%hdl
     end if

     iprio = qrm_starpu_get_prio(tpqrt_, qrm_dscr%pr, prio)

     dscrpt = c_loc(qrm_dscr)
     if(qrm_allocated(a2%stair)) then
        strpt = c_loc(a2%stair(ii))
     else
        strpt = c_null_ptr
     end if

     flops = 2.d0*(real(m)*real(n)*real(n) +      &
                   real(l)*real(l)*real(l)/3.d0 - &
                   real(n)*real(l)*real(l))

     ! write(*,*)'tpqrt  ',m,n,l,flops
     
     call fstarpu_insert_task((/ cqrm_tpqrt_cl,           &
          FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR, &
          FSTARPU_VALUE, c_loc(m),      FSTARPU_SZ_C_INT, &
          FSTARPU_VALUE, c_loc(n),      FSTARPU_SZ_C_INT, &
          FSTARPU_VALUE, c_loc(l),      FSTARPU_SZ_C_INT, &
          FSTARPU_VALUE, c_loc(ib),     FSTARPU_SZ_C_INT, &
          FSTARPU_VALUE, c_loc(ii),     FSTARPU_SZ_C_INT, &
          FSTARPU_VALUE, c_loc(strpt),  FSTARPU_SZ_C_PTR, &
          FSTARPU_RW, hdla1,                              &
          FSTARPU_RW, hdla2,                              &
          FSTARPU_RW, hdlt,                               &
          FSTARPU_SCRATCH, work%hdl,                      &
          FSTARPU_PRIORITY, c_loc(iprio),                 &
          ! FSTARPU_FLOPS, c_loc(flops),                    &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),         &
          C_NULL_PTR /))
     ! call fstarpu_task_wait_for_all()
  else
#endif

     ! make a simple call to lapack tpqrt
     lda1 = size(a1%c,1)
     lda2 = size(a2%c,1)
     ldt  = size(t%c,1)

     __QRM_PROF_START(tb)
     if(qrm_allocated(a2%stair)) then
        call cqrm_tpqrt(m, n, l, ib,    &
             a2%stair(ii),              &
             a1%c(ii,ii), lda1,         &
             a2%c(1,ii), lda2,          &
              t%c(1,ii), ldt,           &
             work%c(1,1), info)
     else
        call cqrm_tpqrt(m, n, l, ib,    &
             -1,                        &
             a1%c(ii,ii), lda1,         &
             a2%c(1,ii), lda2,          &
              t%c(1,ii), ldt,           &
             work%c(1,1), info)
     end if
     __QRM_PROF_STOP(qrm_prof_tpqrt_task_,tb,te)

#if defined (have_starpu)
  end if
#endif

  return
end subroutine cqrm_hitpqrt_task

#if defined (have_starpu)
recursive subroutine cqrm_starpu_tpqrt_cpu_func(buffers, cl_arg) bind(C)
  use qrm_dscr_mod
  use qrm_starpu_common_mod
  use qrm_parameters_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value, intent(in) :: cl_arg
  type(c_ptr), value             :: buffers

  type(c_ptr), target            :: dscr_c, stair_c
  integer(c_int), target         :: i, m, n, l, ib, ii, jj
  integer                        :: ma1, na1, lda1, ma2, na2, lda2, nt, ldt, nw, ldw, info
  complex(r32), pointer          :: a1(:,:), a2(:,:), t(:,:), w(:,:)
  integer, pointer               :: stair(:)
  type(qrm_dscr_type), pointer   :: qrm_dscr
  integer(kind=8)                :: tb, te


  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &
                          c_loc(m),                &
                          c_loc(n),                &
                          c_loc(l),                &
                          c_loc(ib),               &
                          c_loc(ii),               &
                          c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  na1   = fstarpu_matrix_get_ny(buffers, 0)
  lda1  = fstarpu_matrix_get_ld(buffers, 0)

  na2   = fstarpu_matrix_get_ny(buffers, 1)
  lda2  = fstarpu_matrix_get_ld(buffers, 1)

  nt    = fstarpu_matrix_get_ny(buffers, 2)
  ldt   = fstarpu_matrix_get_ld(buffers, 2)

  nw    = fstarpu_matrix_get_ny(buffers, 3)
  ldw   = fstarpu_matrix_get_ld(buffers, 3)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a1, shape=(/lda1,na1/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 1), a2, shape=(/lda2,na2/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 2), t,  shape=(/ldt,nt/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 3), w,  shape=(/ldw,nw/))

  __QRM_PROF_START(tb)
  if(c_associated(stair_c)) then
     call c_f_pointer(stair_c, stair, shape=(/na2/))
     call cqrm_tpqrt(m, n, l, ib,   &
          stair(1),                 &
          a1(ii,1), lda1,           &
          a2(1,1), lda2,            &
           t(1,1), ldt,             &
          w(1,1), info)
  else
     call cqrm_tpqrt(m, n, l, ib,   &
          -1,                       &
          a1(ii,1), lda1,           &
          a2(1,1), lda2,            &
           t(1,1), ldt,             &
          w(1,1), info)
  end if
  __QRM_PROF_STOP(qrm_prof_tpqrt_task_,tb,te)

  return
end subroutine cqrm_starpu_tpqrt_cpu_func

#if defined (have_cuda) && defined (have_cusolver) && defined(have_cuda_qr_panel)
recursive subroutine cqrm_starpu_tpqrt_cuda_func(buffers, cl_arg) bind(C)
  use qrm_dscr_mod
  use qrm_starpu_common_mod
  use qrm_parameters_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value, intent(in) :: cl_arg
  type(c_ptr), value             :: buffers

  type(c_ptr), target            :: dscr_c, stair_c, a1p, a2p, tp, wp, handlep
  integer(c_int), target         :: m, n, l, ib, ii, lw
  integer                        :: ma1, na1, lda1, ma2, na2, lda2, nt, ldt, nw, ldw, info, id
  type(qrm_dscr_type), pointer   :: qrm_dscr
  integer(kind=8)                :: tb, te

  interface
     subroutine cqrm_cuda_tpqrt(         &
          m, n, l, nb,                   &    
          stair,                         &
          a1, lda1,                      &    
          a2, lda2,                      &    
          t , ldt,                       &    
          wd, lwd) bind(C)
       use iso_c_binding
       type(c_ptr), value       :: a1, a2, t, wd, stair
       integer, value           :: m, n, l, nb, lda1, lda2, ldt, lwd
     end subroutine cqrm_cuda_tpqrt
  end interface
  

  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &
                          c_loc(m),                &
                          c_loc(n),                &
                          c_loc(l),                &
                          c_loc(ib),               &
                          c_loc(ii),               &
                          c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  na1   = fstarpu_matrix_get_ny(buffers, 0)
  lda1  = fstarpu_matrix_get_ld(buffers, 0)

  na2   = fstarpu_matrix_get_ny(buffers, 1)
  lda2  = fstarpu_matrix_get_ld(buffers, 1)

  nt    = fstarpu_matrix_get_ny(buffers, 2)
  ldt   = fstarpu_matrix_get_ld(buffers, 2)

  nw    = fstarpu_matrix_get_ny(buffers, 3)
  ldw   = fstarpu_matrix_get_ld(buffers, 3)
  lw    = nw*ldw
  
  a1p = fstarpu_matrix_get_ptr(buffers, 0)
  a2p = fstarpu_matrix_get_ptr(buffers, 1)
  tp  = fstarpu_matrix_get_ptr(buffers, 2)
  wp  = fstarpu_matrix_get_ptr(buffers, 3)

  __QRM_PROF_START(tb)
  call cqrm_cuda_tpqrt(         &
       m, n, l, ib,             &
       stair_c,                 &
       a1p, lda1,               &
       a2p, lda2,               &
       tp, ldt,                 &
       wp, lw)
  __QRM_PROF_STOP(qrm_prof_tpqrt_task_,tb,te)

  return
end subroutine cqrm_starpu_tpqrt_cuda_func
#endif

#endif
