!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT, Inria
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_get_r.F90
!! This file contains a subroutine which extracts the R factor
!! resulting from the factorization in sparse form
!!
!!
!! ##############################################################################################

#include "qrm_common.h"

!> @brief This routine computes the scaled norm of the product A'*r for multiple residuals

!> This routine computes the norm of the scaled product A'*r
!! for a single RHS.
!!
!! @param[in]     qrm_mat a qrm_spmat_type data which contains the R factor to be extracted
!!
!! @param[out]    r  a qrm_spmat_type data which contains the extracted R factor 
!!
subroutine dqrm_spfct_get_schur(qrm_spfct, s, i, j, m, n, info)
  use qrm_common_mod
  use qrm_error_mod
  use dqrm_spmat_mod
  use dqrm_spfct_mod
  use qrm_mem_mod
  implicit none

  type(dqrm_spfct_type), target  :: qrm_spfct
  real(r64)                   :: s(:,:)
  integer, optional              :: i, j, m, n
  integer, optional              :: info

  type(dqrm_dsmat_type), pointer :: f
  integer                        :: is, js, bc, br
  integer                        :: ii, ij, im, in
  integer                        :: bi, bj, bm, bn, bl
  integer                        :: first_rblock, first_cblock
  integer                        :: last_rblock, last_cblock

  ! error management
  integer                          :: err
  character(len=*), parameter      :: name='qrm_get_schur'

  err = 0

  ! mb = qrm_spfct%fdata%front_list(qrm_spfct%adata%schur_node)%f%mb
  f => qrm_spfct%fdata%front_list(qrm_spfct%adata%schur_node)%f

  __QRM_IFPRESENT3(ii, i, 1)
  __QRM_IFPRESENT3(ij, j, 1)
  __QRM_IFPRESENT3(im, m, f%m-ii+1)
  __QRM_IFPRESENT3(in, n, f%n-ij+1)

  
  first_rblock = dqrm_dsmat_inblock(f, ii)
  first_cblock = dqrm_dsmat_inblock(f, ij)
  last_rblock  = dqrm_dsmat_inblock(f, ii+im-1)
  last_cblock  = dqrm_dsmat_inblock(f, ij+in-1)
    
  do bc = first_cblock, last_cblock
     js = max(1,f%f(bc)-ij+1)
     do br = first_rblock, last_rblock
#if defined (have_starpu)
        if(c_associated(f%blocks(br,bc)%hdl)) &
             call fstarpu_data_acquire(f%blocks(br,bc)%hdl, FSTARPU_R)
#endif
        is = max(1,f%f(br)-ii+1)
        call dqrm_dsmat_block_ijmnl(f, i, j, m, n, 0, br, bc, bi, bj, bm, bn, bl)

        
        s(is:is+bm-1,js:js+bn-1) = &
             f%blocks(br,bc)%c(bi:bi+bm-1,bj:bj+bn-1)
#if defined (have_starpu)
        if(c_associated(f%blocks(br,bc)%hdl)) &
             call fstarpu_data_release(f%blocks(br,bc)%hdl)
#endif
     end do
  end do


  if(present(info)) info = err

  return
  
end subroutine dqrm_spfct_get_schur



subroutine dqrm_spfct_get_schur_async(qrm_spfct, hdls, info)
  use qrm_common_mod
  use qrm_error_mod
  use dqrm_spmat_mod
  use dqrm_spfct_mod
  use qrm_mem_mod
  implicit none

  type(dqrm_spfct_type), target  :: qrm_spfct
  type(c_ptr)                    :: hdls(:)
  integer, optional              :: info

  type(dqrm_front_type), pointer :: front
  integer                        :: i, bc, br

  ! error management
  integer                          :: err
  character(len=*), parameter      :: name='qrm_get_schur_async'

  err = 0


  front => qrm_spfct%fdata%front_list(qrm_spfct%adata%schur_node)
  ! mb = front%f%mb
  ! m  = front%f%m
  ! n  = front%f%n

  i = 1

  hdls = c_null_ptr
#if defined (have_starpu)
  if(front%small.eq.0) then
     do bc = 1, front%f%nbc
        do br = 1, front%f%nbr
           if(c_associated(front%f%blocks(br,bc)%hdl)) then
              hdls(i) = front%f%blocks(br,bc)%hdl
              i = i+1
           end if
        end do
     end do
  end if
#endif

  if(present(info)) info = err

  return
  
end subroutine dqrm_spfct_get_schur_async
