!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains the C interface for qr_mumps.
!!
!! ##############################################################################################


#include "qrm_common.h"

!> This module contains the definition of the qr_mumps C interface common
!! to all precisions/types.
module qrm_c_comm_interface
  use iso_c_binding
  use qrm_error_mod
  use qrm_common_mod
  use qrm_string_mod
  use qrm_parameters_mod


  type, bind(c) :: qrm_dscr_type_c
     !> The handle to the qrm_dscr_type instance which will be used
     !! internally by qr_mumps
     type(c_ptr)      :: h
  end type qrm_dscr_type_c

  
contains

  !> C equivalent of the qrm_dscr_init routine
  function qrm_dscr_init_c(qrm_dscr_c) result(info) bind(c)
    use qrm_dscr_mod
    implicit none

    type(qrm_dscr_type_c)        :: qrm_dscr_c
    integer(c_int)               :: info
    type(qrm_dscr_type), pointer :: fdscr

    allocate(fdscr)
    
    call qrm_dscr_init(fdscr)
    qrm_dscr_c%h = c_loc(fdscr)
    nullify(fdscr)

    info = 0
    
    return

  end function qrm_dscr_init_c

  !> C equivalent of the qrm_dscr_destroy routine
  function qrm_dscr_destroy_c(qrm_dscr_c) result(info) bind(c)
    use qrm_dscr_mod
    implicit none

    type(qrm_dscr_type_c)        :: qrm_dscr_c
    integer(c_int)               :: info
    type(qrm_dscr_type), pointer :: fdscr

    call c_f_pointer(qrm_dscr_c%h, fdscr)
    
    call qrm_dscr_destroy(fdscr)
    qrm_dscr_c%h = c_null_ptr

    info = 0
    
    return

  end function qrm_dscr_destroy_c

  !> C equivalent of the qrm_barrier routine
  function qrm_barrier_c() result(info) bind(c)
    use qrm_dscr_mod
    integer :: info
    
    call qrm_barrier()

    info = 0

    return
    
  end function qrm_barrier_c

  !> C equivalent of the qrm_barrier_dscr routine 
  function qrm_barrier_dscr_c(qrm_dscr_c) result(info) bind(c)
    use qrm_dscr_mod
    implicit none
    
    type(qrm_dscr_type_c)        :: qrm_dscr_c
    integer                      :: info
    type(qrm_dscr_type), pointer :: fdscr

    call c_f_pointer(qrm_dscr_c%h, fdscr)
    
    call qrm_barrier(fdscr, info)

    return    
  
  end function qrm_barrier_dscr_c

  
  !> C equivalent of the qrm_init
  !> routine
  function qrm_init_c(ncpu, ngpu) result(info) bind(c)
    use qrm_dscr_mod
    implicit none
    integer(c_int), value  :: ncpu, ngpu
    integer(c_int)         :: info

    call qrm_init(ncpu=ncpu, ngpu=ngpu, info=info)

    return
  end function qrm_init_c
    
  !> C equivalent of the qrm_finalize
  !> routine
  subroutine qrm_finalize_c() bind(c)
    use qrm_dscr_mod
    implicit none

    call qrm_finalize()

    return
  end subroutine qrm_finalize_c

  
  
  
  !> C equivalent of the qrm_glob_set_i4
  !> routine (only for global, integer type)
  function qrm_glob_set_i4_c(string, val) result(info) bind(c)
    implicit none
    
    character(kind=c_char) :: string(*)
    integer(c_int), value  :: val
    integer(c_int)         :: info
    
    character(len=:), allocatable :: a

    a = qrm_c2f_string(string)
    
    call qrm_glob_set_i4(a, val, info)

    return
    
  end function qrm_glob_set_i4_c

  !> C equivalent of the qrm_glob_set_r4
  !> routine (only for global, integer type)
  function qrm_glob_set_r4_c(string, val) result(info) bind(c)
    implicit none

    character(kind=c_char) :: string(*)
    real(c_float), value   :: val
    integer(c_int)         :: info
    
    character(len=:), allocatable :: a

    a = qrm_c2f_string(string)
    
    call qrm_glob_set_r4(a, val, info)

    return
    
  end function qrm_glob_set_r4_c

  !> C equivalent of the qrm_glob_get_i4
  !> routine (only for global, integer type)
  function qrm_glob_get_i4_c(string, val) result(info) bind(c)
    implicit none

    character(kind=c_char) :: string(*)
    integer(c_int)         :: val
    integer(c_int)         :: info
    
    character(len=:), allocatable :: a

    a = qrm_c2f_string(string)
    
    call qrm_glob_get_i4(a, val, info)

    return
    
  end function qrm_glob_get_i4_c

  !> C equivalent of the qrm_glob_get_r4
  !> routine (only for global, real type)
  function qrm_glob_get_r4_c(string, val) result(info) bind(c)
    implicit none

    character(kind=c_char) :: string(*)
    real(c_float)          :: val
    integer(c_int)         :: info
    
    character(len=:), allocatable :: a

    a = qrm_c2f_string(string)
    
    call qrm_glob_get_r4(a, val, info)

    return
    
  end function qrm_glob_get_r4_c

  !> C equivalent of the qrm_glob_get_i8
  !> routine (only for global, integer64 type)
  function qrm_glob_get_i8_c(string, val) result(info) bind(c)
    implicit none

    character(kind=c_char)        :: string(*)
    integer(c_long_long)          :: val
    integer(c_int)                :: info
    
    character(len=:), allocatable :: a

    a = qrm_c2f_string(string)

    call qrm_glob_get_i8(a, val, info)

    return
    
  end function qrm_glob_get_i8_c



  
end module qrm_c_comm_interface

