!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_common_mod.F90
!! this module contains generic interfaces for all the untyped routines.
!!
!!
!! ##############################################################################################

#include "qrm_common.h"

!> @brief This module contains the interfaces of all non-typed routines
module qrm_common_mod

  
  !> @brief Generic interface for the @link ::qrm_print_nsteps_tree @endlink,
  !! @link                                  ::qrm_print_elim_tree @endlink and @link
  !!                                        ::qrm_print_asm_tree @endlink routines
  interface qrm_print_tree
     subroutine qrm_print_nsteps_tree(file, adata, small)
       use qrm_adata_mod
       type(qrm_adata_type)                 :: adata
       character                            :: file*(*)
       integer, optional                    :: small
     end subroutine qrm_print_nsteps_tree
  end interface qrm_print_tree

  !> @brief Generic interface for the @link ::qrm_postorder @endlink routine
  interface qrm_postorder
     subroutine qrm_postorder(parent, n, porder, weight, info)
       integer                              :: n
       integer                              :: parent(:), porder(:)
       integer, optional                    :: weight(:)
       integer, optional                    :: info
     end subroutine qrm_postorder
  end interface qrm_postorder

  !> @brief Generic interface for the @link ::qrm_amalg_tree @endlink routine
  interface
     subroutine qrm_amalg_tree(n, parent, rc, porder, nvar, min_var, &
          fill_thresh, schur_list, schur_size, info)
       use qrm_parameters_mod
       integer                              :: n, min_var, schur_size
       integer                              :: parent(:), rc(:), porder(:), nvar(:), schur_list(:)
       real(r32)                            :: fill_thresh
       integer, optional                    :: info
     end subroutine qrm_amalg_tree
  end interface

  !> @brief Generic interface for the @link ::qrm_compress_data @endlink routine
  interface qrm_compress_data
     subroutine qrm_compress_data(adata, porder, parent, rc, stair, n, info)
       use qrm_mem_mod
       use qrm_adata_mod
       type(qrm_adata_type)                 :: adata
       integer                              :: porder(:), parent(:), rc(:), stair(:)
       integer                              :: n
       integer, optional                    :: info
     end subroutine qrm_compress_data
  end interface qrm_compress_data

  !> @brief Generic interface for the @link ::qrm_reorder_tree @endlink routine
  interface qrm_reorder_tree
     subroutine qrm_reorder_tree(adata, info)
       use qrm_adata_mod
       type(qrm_adata_type)                 :: adata
       integer, optional                    :: info
     end subroutine qrm_reorder_tree
  end interface qrm_reorder_tree

  !> @brief Generic interface for the @link ::qrm_prune_tree @endlink routine
  interface qrm_prune_tree
     subroutine qrm_prune_tree(adata, nth, info)
       use qrm_adata_mod
       type(qrm_adata_type)                 :: adata
       integer                              :: nth
       integer, optional                    :: info
     end subroutine qrm_prune_tree
  end interface qrm_prune_tree

  !> @brief Generic interface for the @link ::qrm_prnt_iarray
  !> @endlink, @link                        ::qrm_prnt_sarray @endlink, @link
  !>                                        ::qrm_prnt_darray @endlink, @link ::qrm_prnt_carray @endlink and
  !> @link                                  ::qrm_prnt_zarray @endlink routines
  interface qrm_prnt_array
     subroutine qrm_prnt_iarray(a, lab, unit)
       integer                              :: a(:)
       character                            :: lab*(*)
       integer, optional                    :: unit
     end subroutine qrm_prnt_iarray
     subroutine qrm_prnt_sarray(a, lab, unit)
       use qrm_parameters_mod
       real(r32)                            :: a(:)
       character                            :: lab*(*)
       integer, optional                    :: unit
     end subroutine qrm_prnt_sarray
     subroutine qrm_prnt_darray(a, lab, unit)
       use qrm_parameters_mod
       real(r64)                            :: a(:)
       character                            :: lab*(*)
       integer, optional                    :: unit
     end subroutine qrm_prnt_darray
     subroutine qrm_prnt_d2array(a, lab, unit)
       use qrm_parameters_mod
       real(r64)                            :: a(:,:)
       character                            :: lab*(*)
       integer, optional                    :: unit
     end subroutine qrm_prnt_d2array
     subroutine qrm_prnt_carray(a, lab, unit)
       use qrm_parameters_mod
       complex(r32)                         :: a(:)
       character                            :: lab*(*)
       integer, optional                    :: unit
     end subroutine qrm_prnt_carray
     subroutine qrm_prnt_zarray(a, lab, unit)
       use qrm_parameters_mod
       complex(r64)                         :: a(:)
       character                            :: lab*(*)
       integer, optional                    :: unit
     end subroutine qrm_prnt_zarray
  end interface qrm_prnt_array

  !> @brief Generic interface for the @link ::qrm_swtime @endlink routine
  interface qrm_swtime
     function qrm_swtime() bind(c, name='qrm_swtime')
       use iso_c_binding
       real(c_double)                     :: qrm_swtime
     end function qrm_swtime
  end interface qrm_swtime

  !> @brief Generic interface for the @link ::qrm_uwtime @endlink routine
  interface qrm_uwtime
     function qrm_uwtime() bind(c, name='qrm_uwtime')
       use iso_c_binding
       real(c_double)                     :: qrm_uwtime
     end function qrm_uwtime
  end interface qrm_uwtime

  !> @brief Generic interface for the @link ::qrm_msleep @endlink routine
  interface qrm_msleep
     subroutine qrm_msleep(n) bind(c, name='qrm_msleep')
       use iso_c_binding
       integer(c_int), value                :: n
     end subroutine qrm_msleep
  end interface qrm_msleep


  !> @brief Generic interface for the @link ::qrm_check_cperm @endlink routine
  interface qrm_check_cperm
     subroutine qrm_check_cperm(cperm, n, info)
       integer                              :: cperm(:)
       integer                              :: n, info
     end subroutine qrm_check_cperm
  end interface qrm_check_cperm

  !> @brief Generic interface for the
  !! @link                                  ::qrm_count_realflops @endlink
  !! @link                                  ::qrm_count_pureflops @endlink
  interface qrm_count_flops
     module procedure qrm_count_realflops, qrm_count_pureflops
  end interface qrm_count_flops

contains



  !> @brief Used for counting the actual flops
  function qrm_count_realflops(m, n, k, op)
    real(kind(1.d0))                        :: qrm_count_realflops
    integer                                 :: m, k, n
    character                               :: op*(*)

    real(kind(1.d0))                        :: rk, rm, rn

    rk = real(k, kind(1.d0))
    rm = real(m, kind(1.d0))
    rn = real(n, kind(1.d0))

    qrm_count_realflops = 0.d0

    select case(op)
    case('panel')
       if(m .ge. k) then
          qrm_count_realflops = (2.d0*rk*rk*(rm-rk/3.d0))
       else
          rk = rm
          rn = real(n-m, kind(1.d0))
          qrm_count_realflops = (2.d0*rm*rm*(rm-rm/3.d0)) + (rk*rn*(4.d0*rm-rk))
       end if
    case('update')
       qrm_count_realflops = rk*rn*(4.d0*rm-rk)
    end select

    return
  end function qrm_count_realflops


  !> @brief Used for counting the real flops (i.e., it ignores the zeros
  !> included by the blocking)
  function qrm_count_pureflops(stair, n, j, nb)
    implicit none
    real(kind(1.d0))                        :: qrm_count_pureflops
    integer                                 :: stair(:)
    integer                                 :: n, nb, j

    integer                                 :: i

    qrm_count_pureflops=0
    do i=j, min(j+nb-1, n)
       qrm_count_pureflops = qrm_count_pureflops+(stair(i)-i+1)*(3 + 4*(n-i))
    end do

    return
  end function qrm_count_pureflops


  subroutine sleep(t)
    implicit none
    integer :: t_start, t_end, t_rate
    real(kind(1.d0))     :: ta, t
    call system_clock(t_start)
    do
       call system_clock(t_end, t_rate)
       ta = real(t_end-t_start)/real(t_rate)    
       if(ta.gt.t) return
    end do
  end subroutine sleep




  function qrm_inblock(f, i) result(b)
    implicit none
    
    integer    :: f(:)
    integer    :: i, b
    
    integer    :: nb, k
    
    b = 0
    
    if(i.gt.f(size(f))) return
    
    do b=1, size(f)
       if((i.ge. f(b)) .and. (i.lt. f(b+1))) exit
    end do
    
    return

  end function qrm_inblock



  ! Assume a matrix partitioned into blocks of variable size whose&
  ! first indices are given in a%f. We want to compute what part of the
  ! (br,bc) block overlaps with the trapezoidal submatrix described by
  ! i, j, m, n, l. This is given by bi, bj, bm, bn ,bl.
subroutine qrm_block_ijmnl(f, i, j, m, n, l, br, bc, bi, bj, bm, bn, bl)

  implicit none

  integer    :: f(:)
  integer    :: i, j, m, n, l, br, bc, bi, bj, bm, bn, bl

  integer    :: ii, jj, t

  ii = max(i,f(br))
  jj = max(j,f(bc))
  bn = min(f(bc+1),j+n)-max(j,f(bc))
  bm = min(f(br+1),i+m)-max(i,f(br))
  ! write(*,*)i,j,m,n,l,ii,jj,bm,bn
  if (l.gt.0) then
     t  = i+m-l+jj-j
     if(t.lt.ii) then
        jj = jj+(ii-t)
        bn = bn-(ii-t)
        bm = min(bm,bn)
        bl = bm
     else
        bl = max(0,min(f(br+1),i+m)-t)
     end if
  else if (l.lt.0) then
     t  = j+n+l+ii-i
     if(t.lt.jj) then
        ii = ii+(jj-t)
        bm = bm-(jj-t)
        bn = min(bm,bn)
        bl = -bm
     else
        bl = -max(0,min(f(bc+1),j+n)-t)
     end if
  else
     bl = 0
  end if
  bj = max(jj-f(bc)+1, 1)
  bi = max(ii-f(br)+1, 1)

  return
end subroutine qrm_block_ijmnl



subroutine qrm_build_fc(fc, m, n, nbr, nbc, mb, mbs, npiv, fc_in)
  use qrm_mem_mod
  implicit none

  integer, allocatable :: fc(:)
  integer              :: m, n, nbr, nbc
  integer, optional    :: mb, npiv, mbs(:), fc_in(:)

  integer              :: nb, inpiv, mx, bs, b

  mx  = max(m,n)
  nbr = -1
  nbc = -1

  if(present(mb)) then
     nb = (mx-1)/mb+1
     if(present(npiv)) then
        inpiv = npiv
        nb = nb+1
     else
        inpiv = mx+1
     end if
     
     call qrm_alloc(fc, nb+1)
     fc(1) = 1
     nb = 1
     
     do
        bs = min(mb,    mx-fc(nb)+1)
        if(fc(nb).le.inpiv) bs = min(bs, inpiv-fc(nb)+1)

        fc(nb+1) = fc(nb)+bs

        if(fc(nb+1).gt.m .and. nbr.eq.-1) nbr = nb
        if(fc(nb+1).gt.n .and. nbc.eq.-1) nbc = nb
        if(fc(nb+1).gt.mx) then
           exit
        else
           nb = nb+1
        end if
     end do

     do b=nb+2, size(fc)
        fc(b)=fc(nb+1)
     end do

  else if (present(mbs)) then
     nb = size(mbs)+1

     call qrm_alloc(fc, nb)
     fc(1) = 1
     nb = 1

     do
        bs = min(mbs(nb), mx-fc(nb)+1)
        
        fc(nb+1) = fc(nb)+bs
        if(fc(nb+1).gt.m .and. nbr.eq.-1) nbr = nb
        if(fc(nb+1).gt.n .and. nbc.eq.-1) nbc = nb
        if(fc(nb+1).gt.mx) then
           exit
        else
           nb = nb+1
        end if
     end do

     do b=nb+2, size(fc)
        fc(b)=fc(nb+1)
     end do

  else if (present(fc_in)) then
     nb = size(fc_in)
     
     call qrm_alloc(fc, nb)
     fc(1) = 1
     nb = 1
     
     do
        bs = min(fc_in(nb+1)-fc_in(nb), mx-fc(nb)+1)

        fc(nb+1) = fc(nb)+bs
        if(fc(nb+1).gt.m .and. nbr.eq.-1) nbr = nb
        if(fc(nb+1).gt.n .and. nbc.eq.-1) nbc = nb
        if(fc(nb+1).gt.mx) then
           exit
        else
           nb = nb+1
        end if
     end do

     do b=nb+2, size(fc)
        fc(b)=fc(nb+1)
     end do
  else
     if(present(npiv)) then
        call qrm_alloc(fc, 3)
        fc(1) = 1
        fc(2) = min(mx+1,npiv+1)
        fc(3) = mx+1
        if(npiv.lt.m) then
           nbr = 2
        else
           nbr = 1
        end if
        if(npiv.lt.n) then
           nbc = 2
        else
           nbc = 1
        end if
     else
        call qrm_alloc(fc, 2)
        fc(1) = 1
        fc(2) = mx+1
        nbr   = 1
        nbc   = 1
     end if
  end if
  
  return
  
end subroutine qrm_build_fc


  
end module qrm_common_mod
