!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!!##############################################################################################
!> @file dqrm_spgeqr.F90
!!
!! An example program which reads parameters from a file, a matrix from a MM file and
!! solves the problem. Run this program as
!!
!! ./dqrm_spgeqr -mat matrix -nrhs nr
!!
!! (where _ is d, s, c or z) using the provided input.txt file.
!!
!!
!!##############################################################################################

#include "qrm_common.h"

program dqrm_spgeqr

  use dqrm_mod
  implicit none

  type(dqrm_spmat_type)          :: qrm_spmat
  type(dqrm_spfct_type)          :: qrm_spfct

  integer                        :: info, i, nrhs, ounit, n, verb
  integer                        :: mb, nb, ib, bh, imx, idamax
  character                      :: transp 
  real(r64), allocatable      :: b(:,:), x(:,:), r(:,:)
  real(kind(1.d0))               :: tanl, tfct, tslv
  integer(i64)                   :: ts, te, tr, nnzr, nnzh, gflops
  real(r64)                      :: anrm, mxrnrm, mxxnrm, mxonrm, mxbnrm
  real(r64), allocatable         :: rnrm(:), onrm(:), bnrm(:), xnrm(:)
  character(len=:), allocatable  :: matname
  integer                        :: ncpu, ngpu
  
  call qrm_init()
  call qrm_stop_profiling()

  ! initialize the matrix data structure. 
  call qrm_spmat_init(qrm_spmat)
  call qrm_stop_profiling()

  ! read the input file
  call qrm_read_parms(matname, qrm_spmat, nrhs, verb)

  call qrm_spfct_init(qrm_spfct, qrm_spmat)
  
  ounit = 0
  call qrm_get('qrm_ounit', ounit)
  call qrm_get('qrm_mb', mb)
  call qrm_get('qrm_ncpu', ncpu)
  call qrm_get('qrm_ngpu', ngpu)

  call qrm_log_record("Setup", verb=2)
  call qrm_log_record("- Matrix name","mat","a20",matname)
  call qrm_log_record("- M","m","i10",qrm_spmat%m, verb=2)
  call qrm_log_record("- N","n","i10",qrm_spmat%n, verb=2)
  call qrm_log_record("- NNZ","nnz","i15",qrm_spmat%nz, verb=2)
  call qrm_log_record("- MB","mb","i4",mb)
  call qrm_log_record("- NCPU","ncpu","i4",ncpu)
  call qrm_log_record("- NGPU","ngpu","i4",ngpu)
  if(qrm_spmat%m .ge. qrm_spmat%n) then
     transp=qrm_no_transp
  else
     transp=qrm_transp
  end if

  if(verb.ge.2) call qrm_log_print('list',verb=verb)


  call qrm_log_record("Analysis", verb=2)
  call system_clock(ts)
  call qrm_analyse(qrm_spmat, qrm_spfct, transp, info)
  call system_clock(te, tr); tanl = real(te-ts)/real(tr)
  __QRM_INFO_CHECK(info,'qrm_test','qrm_analyse',10)
  
  gflops = qrm_spfct%gstats(qrm_e_facto_flops_)

  call qrm_log_record("- Estimated nonzeroes in R","e-nnz-r","i20",qrm_spfct%gstats(qrm_e_nnz_r_), verb=2)
  call qrm_log_record("- Estimated nonzeroes in H","e-nnz-h","i20",qrm_spfct%gstats(qrm_e_nnz_h_), verb=2)
  call qrm_log_record("- Estimated factorization flops","e-f-flops","i20",gflops, verb=2)
  call qrm_log_record("- Estimated factorization mem. peak (MB)","e-mem","f20.3", &
                       real(qrm_spfct%gstats(qrm_e_facto_mempeak_))/1e6, verb=2)
  call qrm_log_record("- Time for analysis","Ta","es10.3",tanl)
  if(verb.ge.2) call qrm_log_print('list',verb=verb)

  call qrm_log_record("Factorization", verb=2)
  call qrm_start_profiling()
  call system_clock(ts)
  call qrm_factorize(qrm_spmat, qrm_spfct, transp, info)
  call system_clock(te, tr); tfct = real(te-ts)/real(tr)
  call qrm_stop_profiling()
  __QRM_INFO_CHECK(info,'qrm_test','qrm_factorize',10)

  call qrm_log_record("- Actual NNZ in R","r-nnz","i20",qrm_spfct%gstats(qrm_nnz_r_), verb=2)
  call qrm_log_record("- Gflops/s at facto","f-gflops","f8.2",real(gflops)/tfct/1e9)
  call qrm_log_record("- Time for factorization","Tf","es10.3",tfct)
  if(verb.ge.2) call qrm_log_print('list',verb=verb)

  if((qrm_spfct%icntl(qrm_keeph_) .eq. 1) .and. (nrhs.gt.0)) then
     call qrm_log_record("Solve", verb=2)
     call qrm_alloc(b , qrm_spmat%m, nrhs)
     call qrm_alloc(r , qrm_spmat%m, nrhs)
     call qrm_alloc(x , qrm_spmat%n, nrhs)
     
     call qrm_alloc(bnrm, nrhs)
     call qrm_alloc(rnrm, nrhs)
     call qrm_alloc(xnrm, nrhs)
     call qrm_alloc(onrm, nrhs)

     bnrm = qrm_dzero 
     rnrm = qrm_dzero 
     xnrm = qrm_dzero 
     onrm = qrm_dzero 
     call qrm_vec_larnv(b)

     x  = qrm_dzero
     r  = b

     call system_clock(ts)
     call qrm_spfct_geqrs(qrm_spfct, b, x, transp)
     call system_clock(te, tr); tslv = real(te-ts)/real(tr)

     ! compute the residual
     call qrm_residual_norm(qrm_spmat, r, x, rnrm)

     call qrm_vecnrm(x, size(x,1), '2', xnrm)
     call qrm_vecnrm(b, size(b,1), '2', bnrm)
     call qrm_spmat_nrm(qrm_spmat, 'f', anrm)
     call qrm_residual_orth(qrm_spmat, r, onrm)   

     if(qrm_spmat%m.gt.qrm_spmat%n) then
        imx = idamax(nrhs, onrm, 1)
     else
        imx = idamax(nrhs, rnrm, 1)
     end if
     
     mxxnrm = xnrm(imx)
     mxrnrm = rnrm(imx)
     mxbnrm = bnrm(imx)
     mxonrm = onrm(imx)
     
     call qrm_log_record("- A norm","nrm-a","es10.3",anrm)
     call qrm_log_record("- Max residual norm","nrm-r","es10.3",mxrnrm)
     call qrm_log_record("- Max orth-residual norm","nrm-o","es10.3",mxonrm)
     call qrm_log_record("- Max x norm","nrm-x","es10.3",mxxnrm)
     call qrm_log_record("- Max b norm","nrm-b","es10.3",mxbnrm)
     call qrm_log_record("- Time for solve","Ts","es10.3",tslv)
     if(verb.ge.2) call qrm_log_print('list',verb=verb)
     
     call qrm_dealloc(b)
     call qrm_dealloc(r)
     call qrm_dealloc(x)
     call qrm_dealloc(bnrm)
     call qrm_dealloc(rnrm)
     call qrm_dealloc(xnrm)
     call qrm_dealloc(onrm)
  else
     tslv = 0.0
     mxxnrm = 0.0
     mxrnrm = 0.0
     mxonrm = 0.0
  end if

10 continue

  call qrm_spmat_destroy(qrm_spmat)
  call qrm_spfct_destroy(qrm_spfct)

  call qrm_log_record("Finalization", verb=2)
  call qrm_log_record("- Total time","Tt","es10.3",tanl+tfct+tslv, verb=1)
  call qrm_log_record("- Total unallocated memory","u-mem","i20",qrm_tot_mem, verb=2)
  call qrm_log_record("- Actual memory peak","p-mem","f9.3",real(qrm_max_mem,r32)/1e6)
  call qrm_log_record("=", verb=2)
  if(verb.ge.2) call qrm_log_print('list',verb=verb)

  if(verb.lt.2) call qrm_log_print('tab',verb=verb)

  call qrm_finalize()
  stop

contains

  subroutine qrm_read_parms(matname, qrm_spmat, nrhs, verb)
    ! Function: read_parms
    !
    ! *Input*:
    !
    ! *Output*:
    !
    use qrm_ssget_mod
    use dqrm_utils_mod
    use qrm_cli_mod
    implicit none

    type(dqrm_spmat_type)         :: qrm_spmat
    integer                       :: nrhs
    integer                       :: verb
    character(len=:), allocatable :: matname

    type(qrm_cli_type)            :: cli
    character(len=:), allocatable :: mat, matfile, ssrepo, ssmat
    integer                       :: col1, col2, col3, col4, err
    integer                       :: nx, ny, nz, n, s
    real(r32)                     :: gamma

    call cli%init(title='dqrm_spgeqr', description='A test program for the QR factorization of sparse matrices')
    call cli%add('-mat'  , 'The matrix file/name.\n'//&
         'This can either be a file in MatrixMarket format, \n'//&
         'a string of the form ss:path:matname, in which case \nmatrix '//&
         'matname from SuiteSparse will be downloaded \nin path and used '//&
         'or a string of the type Mitt3D_27:N \nwhere a Mitt3D_27 matrix of '//&
         'size N will be generated.', def='', req=.true.)
    call cli%add('-nrhs' , 'The number of right-hand sides.'                            , def=1 )
    call cli%add('-v'    , 'Verbosity: 3=full, 2=batch w/ header, 1=batch w/o header.'  , def=3 )
    call cli%add('-t'    , 'Apply Tikhonov regularization.'                             , def=0.0_r32)

    call cli%parse(info)
    if(info.ne.0) then
       stop
    end if
    
    call cli%get('-mat'  , mat)
    call cli%get('-nrhs' , nrhs)
    call cli%get('-v'    , verb)
    call cli%get('-t'    , gamma)

    if(index(mat, 'ss:').eq.1) then
       ! this is a SuiteSparse matrix
       col1 = index(mat, ':')
       if(col1.eq.0) then
          write(*,'("SS matrix not correctly formatted.")')
          stop
       end if
       col2 = index(mat(col1+1:),':')+col1
       if(col2.eq.col1 .or. col2.eq.col1+1) then
          ssrepo = '.'
       else
          ssrepo = mat(col1+1:col2-1)
       end if
       call qrm_ssget_init(ssrepo)
       matfile = qrm_ssget(mat(col2+1:))
       call qrm_readmat(matfile, qrm_spmat, .true., err)
       __QRM_INFO_CHECK(err,'qrm_test','qrm_readmat',20)
       matname = mat(col2+1:)
       ! else if(index(mat, '7PT:').eq.1) then
       ! this is a 7PT matrix
       ! col1 = index(mat, ':')
       ! col2 = index(mat(col1+1:),':')+col1
       ! col3 = index(mat(col2+1:),':')+col2
       ! col4 = index(mat(col3+1:),':')+col3
       ! if(col1.eq.0 .or. col2.eq.col1 .or. col3.eq.col2 .or. col4.eq.col3) then
          ! write(*,'("7PT matrix not correctly formatted.")')
          ! stop
       ! end if
       ! read(mat(col1+1:col2-1),*)nx
       ! read(mat(col2+1:col3-1),*)ny
       ! read(mat(col3+1:col4-1),*)nz
       ! read(mat(col4+1:      ),*)s
       ! call qrm_matgen_7pt(qrm_spmat, nx, ny, nz, s)
    else if(index(mat, 'Mitt3D_27:').eq.1) then
       ! this is a Mitt3D matrix
       col1 = index(mat, ':')
       if(col1.eq.0) then
          write(*,'("Mitt3D_27 matrix not correctly formatted.")')
          stop
       end if
       read(mat(col1+1:),*)n
       call qrm_matgen_mitt3d(qrm_spmat, n)
       matname = mat
    else
       matfile = mat
       call qrm_readmat(matfile, qrm_spmat, .true., err)
       __QRM_INFO_CHECK(err,'qrm_test','qrm_readmat',20) 
       matname = mat
    end if

    if(gamma.ne.0.0) then
       call qrm_tikhonov(qrm_spmat, real(gamma,kind(qrm_spmat%val(1))))
    end if

    ounit = 0
    if(verb.eq.3) call qrm_get('qrm_ounit', ounit)

20  continue
    ! if(present(info)) info = err
    return
    
  end subroutine qrm_read_parms


  
end program dqrm_spgeqr






